<?php
/**
 * Plugin Name: Filerobot - Digital Asset Management, Media Optimisation and Acceleration
 * Description: Filerobot by Scaleflex normalizes, resizes, optimizes and distributes your images rocket fast around the world.
 * Version: 4.0.3
 * Requires PHP: 5.3.3
 * Requires at least: 4.8
 * Author: Scaleflex
 * Author URI: https://www.scaleflex.com/en/home
 * License:           GPL-2.0+
 * License URI:       http://www.gnu.org/licenses/gpl-2.0.txt
 * Text Domain: filerobot
 * Domain Path: /languages
 */

$plugin = plugin_basename(__FILE__); 
load_plugin_textdomain('filerobot', false, dirname($plugin) . '/languages/');

$filerobot_class = dirname(__FILE__) . DIRECTORY_SEPARATOR . 'filerobot_class.php';
$filerobot_api = dirname(__FILE__) . DIRECTORY_SEPARATOR . 'filerobot_api.php';

function filerobot_plugin_settings_link($links) 
{ 
    $url = esc_url( add_query_arg(
        'page',
        'filerobot',
        get_admin_url() . 'admin.php'
    ) ) . '&tab=settings';

    $settings_link = "<a href='$url'>" . __( 'Settings' ) . '</a>';

    array_unshift(
        $links,
        $settings_link
    );
    return $links;
}
add_filter("plugin_action_links_$plugin", 'filerobot_plugin_settings_link' );

require $filerobot_class;
require $filerobot_api;

function filerobot_incompatibile($msg)
{
    require_once ABSPATH . DIRECTORY_SEPARATOR . 'wp-admin' . DIRECTORY_SEPARATOR . 'includes' . DIRECTORY_SEPARATOR . 'plugin.php';
    deactivate_plugins(__FILE__);

    wp_die($msg);
}

if (is_admin() && (!defined('DOING_AJAX') || !DOING_AJAX)) 
{
    if (version_compare(PHP_VERSION, '5.3.3', '<')) 
    {
        filerobot_incompatibile( 'Plugin Filerobot Spaces Sync requires PHP 5.3.3 or higher. The plugin has now disabled itself.' );

    } 
    elseif ( !function_exists('curl_version') || !($curl = curl_version()) || empty($curl['version']) || empty($curl['features']) || version_compare($curl['version'], '7.16.2', '<') ) 
    {
        filerobot_incompatibile( 'Plugin Filerobot Spaces Sync requires cURL 7.16.2+. The plugin has now disabled itself.' );
    } 
    elseif (!($curl['features'] & CURL_VERSION_SSL)) 
    {
        filerobot_incompatibile( 'Plugin Filerobot Spaces Sync requires that cURL is compiled with OpenSSL. The plugin has now disabled itself.' );
    }
}

function run_filerobot()
{
    $instance = new Filerobot();
    $instance->setup();
}

function on_activation()
{
    Filerobot::setup_db();
}
function on_deactivation()
{
    Filerobot::deactivate();
}
// function on_uninstall()
// {
//     Filerobot::uninstall();
// }

register_activation_hook( __FILE__, "on_activation" );
register_deactivation_hook(__FILE__, 'on_deactivation');
// register_uninstall_hook(__FILE__, 'on_uninstall');

run_filerobot();
